/***********************************************************************/
/*                                                                     */
/*                               Pio (C)                               */
/*                                                                     */
/*                    Polyglot Input/Output Library                    */
/*                                                                     */
/*               EPI Pomdapi, INRIA Paris - Rocquencourt               */
/*                                                                     */
/*  Copyright 2011-2014 INRIA.                                         */
/*  Institut National de Recherche en Informatique et en Automatique.  */
/*  All rights reserved. This file is distributed only by permission.  */
/*                                                                     */
/*  Bruno Verlyck <Bruno.Verlyck@inria.fr>                             */
/*                                                                     */
/***********************************************************************/

#ifndef TYPED_VALUE_INCLUDED
#define TYPED_VALUE_INCLUDED

#include "value_type_def.h"
#include <stdlib.h>

typedef enum {
  /* type scalar */
  TV_Sbool = 1,				/* catch zero'd memory */
  TV_Sbool_couple,
  TV_Sbool_triple,
  TV_Sstring,
  TV_Sint,
  TV_Sint_couple,
  TV_Sint_triple,
  TV_Snative_int,
  TV_Snative_int_couple,
  TV_Snative_int_triple,
  TV_Sint32,
  TV_Sint32_couple,
  TV_Sint32_triple,
  TV_Sint64,
  TV_Sint64_couple,
  TV_Sint64_triple,
  TV_Sfloat,
  TV_Sfloat_couple,
  TV_Sfloat_triple,
  /* type vector */
  TV_Vbool,
  TV_Vbool_couple,
  TV_Vbool_triple,
  TV_Vstring,
  TV_Vint,
  TV_Vint_couple,
  TV_Vint_triple,
  TV_Vnative_int,
  TV_Vnative_int_couple,
  TV_Vnative_int_triple,
  TV_Vint32,
  TV_Vint32_couple,
  TV_Vint32_triple,
  TV_Vint64,
  TV_Vint64_couple,
  TV_Vint64_triple,
  TV_Vfloat,
  TV_Vfloat_couple,
  TV_Vfloat_triple,
  /* type matrix */
  TV_Mc_float,
  TV_Mfortran_float,
} TV_tag;

typedef struct {
  TV_tag tv_tag;
  union {
    bool b;					    /* TV_Sbool */
    struct cbool { bool b1, b2; } cb;		    /* TV_Sbool_couple */
    struct tbool { bool b1, b2, b3; } tb;	    /* TV_Sbool_triple */
    struct string str;				    /* TV_Sstring */
    int i;					    /* TV_Sint */
    struct cint { int i1, i2; } ci;		    /* TV_Sint_couple */
    struct tint { int i1, i2, i3; } ti;		    /* TV_Sint_triple */
    natint nat;					    /* TV_Snative_int */
    struct cnat { natint nat1, nat2; } cnat;	    /* TV_Snative_int_couple */
    struct tnat { natint nat1, nat2, nat3; } tnat;  /* TV_Snative_int_triple */
    int32 i32;					    /* TV_Sint32 */
    struct cint32 { int32 i321, i322; } ci32;	    /* TV_Sint32_couple */
    struct tint32 { int32 i321, i322, i323; } ti32; /* TV_Sint32_triple */
    int64 i64;					    /* TV_Sint64 */
    struct cint64 { int64 i641, i642; } ci64;	    /* TV_Sint64_couple */
    struct tint64 { int64 i641, i642, i643; } ti64; /* TV_Sint64_triple */
    double f;					    /* TV_Sfloat */
    struct cfloat { double f1, f2; } cf;	    /* TV_Sfloat_couple */
    struct tfloat { double f1, f2, f3; } tf;	    /* TV_Sfloat_triple */
    struct vector {
      size_t vsize;
      union {
	bool *vb;			/* TV_Vbool */
	struct cbool *vcb;		/* TV_Vbool_couple */
	struct tbool *vtb;		/* TV_Vbool_triple */
	struct string *vs;		/* TV_Vstring */
	int *vi;			/* TV_Vint */
	struct cint *vci;		/* TV_Vint_couple */
	struct tint *vti;		/* TV_Vint_triple */
	natint *vnat;			/* TV_Vnative_int */
	struct cnat *vcnat;		/* TV_Vnative_int_couple */
	struct tnat *vtnat;		/* TV_Vnative_int_triple */
	int32 *vi32;			/* TV_Vint32 */
	struct cint32 *vci32;		/* TV_Vint32_couple */
	struct tint32 *vti32;		/* TV_Vint32_triple */
	int64 *vi64;			/* TV_Vint64 */
	struct cint64 *vci64;		/* TV_Vint64_couple */
	struct tint64 *vti64;		/* TV_Vint64_triple */
	double *vf;			/* TV_Vfloat */
	struct cfloat *vcf;		/* TV_Vfloat_couple */
	struct tfloat *vtf;		/* TV_Vfloat_triple */
	void *vv;			/* needed in read_typed_value.c */
      } p;				/* as in 'ptr' */
    } vec;
    struct matrix {			/* TV_Mcfloat */
      size_t msize1, msize2;		/* TV_Mffloat */
      double *mf;			/* an union is useless here */
    } mat;
  } v;					/* as in 'value' */
} TV_t;

vtd_value_type type_of_value (TV_t *v);

/* exception Wrong_communication of string */
/* ( should only be ?) catched by read_communication.c */
jmp_buf *wrong_communication_label;
char *wrong_communication_msg;
void raise_wrong_communication (char *msg) __attribute__ ((noreturn));
void raise_wrong_communication (char *msg);

/* This is to use the same macro calls for these declarations and
   typed_value.c's implementations, where scalarity and fieldnam are used */
#define TV_TO(scalarity, what, vectorp, rettype, fieldnam) \
rettype to_##what##vectorp (TV_t *typed_value);

/* These 3 lines are a small sanity check */
bool to_bool (TV_t *typed_value);
struct cbool to_bool_couple (TV_t *typed_value);
struct tbool to_bool_triple (TV_t *typed_value);
TV_TO (S, bool, , bool, b)
TV_TO (S, bool_couple, , struct cbool, cb)
TV_TO (S, bool_triple, , struct tbool, tb)
TV_TO (S, string, , struct string, str)
TV_TO (S, int, , int, i)
TV_TO (S, int_couple, , struct cint, ci)
TV_TO (S, int_triple, , struct tint, ti)
TV_TO (S, native_int, , natint, nat)
TV_TO (S, native_int_couple, , struct cnat, cnat)
TV_TO (S, native_int_triple, , struct tnat, tnat)
TV_TO (S, int32, , int32, i32)
TV_TO (S, int32_couple, , struct cint32, ci32)
TV_TO (S, int32_triple, , struct tint32, ti32)
TV_TO (S, int64, , int64, i64)
TV_TO (S, int64_couple, , struct cint64, ci64)
TV_TO (S, int64_triple, , struct tint64, ti64)
TV_TO (S, float, , double, f)
TV_TO (S, float_couple, , struct cfloat, cf)
TV_TO (S, float_triple, , struct tfloat, tf)

/* Another sanity check */
_Bool * to_bool_vector (TV_t *typed_value);
struct cbool * to_bool_couple_vector (TV_t *typed_value);
struct tbool * to_bool_triple_vector (TV_t *typed_value);
TV_TO (V, bool, _vector, bool *, vb)
TV_TO (V, bool_couple, _vector, struct cbool *, vcb)
TV_TO (V, bool_triple, _vector, struct tbool *, vtb)
TV_TO (V, string, _vector, struct string *, vec.vs)
TV_TO (V, int, _vector, int *, vi)
TV_TO (V, int_couple, _vector, struct cint *, vci)
TV_TO (V, int_triple, _vector, struct tint *, vti)
TV_TO (V, native_int, _vector, natint *, vnat)
TV_TO (V, native_int_couple, _vector, struct cnat *, vcnat)
TV_TO (V, native_int_triple, _vector, struct tnat *, vtnat)
TV_TO (V, int32, _vector, int32 *, vi32)
TV_TO (V, int32_couple, _vector, struct cint32 *, vci32)
TV_TO (V, int32_triple, _vector, struct tint32 *, vti32)
TV_TO (V, int64, _vector, int64 *, vi64)
TV_TO (V, int64_couple, _vector, struct cint64 *, vci64)
TV_TO (V, int64_triple, _vector, struct tint64 *, vti64)
TV_TO (V, float, _vector, double *, vf)
TV_TO (V, float_couple, _vector, struct cfloat *, vcf)
TV_TO (V, float_triple, _vector, struct tfloat *, vtf)

TV_TO (M, c_float, _matrix, double *, mat.mf)
TV_TO (M, fortran_float, _matrix, double *, mat.mf)
#undef TV_TO

/* By value version */
#define TV_of_V(kind, type, fieldnam)		\
TV_t *TV_of_##kind (type val);

/* By reference version */
#define TV_of_R(kind, type, fieldnam)		\
TV_t *TV_of_##kind (type *val);

TV_of_V (bool, bool, b)
TV_of_R (bool_couple, struct cbool, cb)
TV_of_R (bool_triple, struct tbool, tb)
TV_of_R (string, struct string, str)
TV_of_V (int, int, i)
TV_of_R (int_couple, struct cint, ci)
TV_of_R (int_triple, struct tint, ti)
TV_of_V (native_int, natint, nat)
TV_of_R (native_int_couple, struct cnat, cnat)
TV_of_R (native_int_triple, struct tnat, tnat)
TV_of_V (int32, int32, i32)
TV_of_R (int32_couple, struct cint32, ci32)
TV_of_R (int32_triple, struct tint32, ti32)
TV_of_V (int64, int64, i64)
TV_of_R (int64_couple, struct cint64, ci64)
TV_of_R (int64_triple, struct tint64, ti64)
TV_of_V (float, double, f)
TV_of_R (float_couple, struct cfloat, cf)
TV_of_R (float_triple, struct tfloat, tf)

#undef TV_of_V
#undef TV_of_R

#define TV_of_vec(kind, type)				\
  TV_t *TV_of_##kind##_vector (struct vector v);

TV_of_vec (bool, bool)
TV_of_vec (bool_couple, struct cbool)
TV_of_vec (bool_triple, struct tbool)
TV_of_vec (string, struct string)
TV_of_vec (int, int)
TV_of_vec (int_couple, struct cint)
TV_of_vec (int_triple, struct tint)
TV_of_vec (native_int, natint)
TV_of_vec (native_int_couple, struct cnat)
TV_of_vec (native_int_triple, struct tnat)
TV_of_vec (int32, int32)
TV_of_vec (int32_couple, struct cint32)
TV_of_vec (int32_triple, struct tint32)
TV_of_vec (int64, int64)
TV_of_vec (int64_couple, struct cint64)
TV_of_vec (int64_triple, struct tint64)
TV_of_vec (float, double)
TV_of_vec (float_couple, struct cfloat)
TV_of_vec (float_triple, struct tfloat)

/* Convenience funcs for the mundane C programmer */
TV_t *TV_of_chars (const char *s);
TV_t *TV_of_chars_array (const char **s, size_t asiz);

#undef TV_of_vec

TV_t *TV_of_c_float_matrix (double *m, size_t sizei, size_t sizej);
TV_t *TV_of_fortran_float_matrix (double *m, size_t sizei, size_t sizej);

#endif
